'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.joinPrefix = joinPrefix;
exports.quotePrefix = quotePrefix;
exports.thisIsNotTheEndOfThisBatch = thisIsNotTheEndOfThisBatch;
exports.whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch = whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch;
exports.keysetPagingSelect = keysetPagingSelect;
exports.offsetPagingSelect = offsetPagingSelect;
exports.orderColumnsToString = orderColumnsToString;
exports.interpretForOffsetPaging = interpretForOffsetPaging;
exports.interpretForKeysetPaging = interpretForKeysetPaging;
exports.validateCursor = validateCursor;

var _lodash = require('lodash');

var _graphqlRelay = require('graphql-relay');

var _util = require('../util');

function joinPrefix(prefix) {
  return prefix.slice(1).map(name => name + '__').join('');
}

function doubleQuote(str) {
  return `"${str}"`;
}

function quotePrefix(prefix, q = doubleQuote) {
  return prefix.map(name => q(name));
}

function thisIsNotTheEndOfThisBatch(node, parent) {
  return !node.sqlBatch && !node.junctionBatch || !parent;
}

function whereConditionIsntSupposedToGoInsideSubqueryOrOnNextBatch(node, parent) {
  return !node.paginate && (!node.sqlBatch || !parent);
}

function keysetPagingSelect(table, whereCondition, orderColumns, limit, as, options = {}) {
  let { joinCondition, joinType, q } = options;
  q = q || doubleQuote;
  whereCondition = (0, _lodash.filter)(whereCondition).join(' AND ') || 'TRUE';
  if (joinCondition) {
    return `\
${joinType || ''} JOIN LATERAL (
  SELECT * FROM ${table}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns, q)}
  LIMIT ${limit}
) ${q(as)} ON ${joinCondition}`;
  } else {
    return `\
FROM (
  SELECT * FROM ${table}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns, q)}
  LIMIT ${limit}
) ${q(as)}`;
  }
}

function offsetPagingSelect(table, pagingWhereConditions, orderColumns, limit, offset, as, options = {}) {
  let { joinCondition, joinType, q } = options;
  q = q || doubleQuote;
  const whereCondition = (0, _lodash.filter)(pagingWhereConditions).join(' AND ') || 'TRUE';
  if (joinCondition) {
    return `\
${joinType || ''} JOIN LATERAL (
  SELECT *, count(*) OVER () AS ${q('$total')}
  FROM ${table}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns, q)}
  LIMIT ${limit} OFFSET ${offset}
) ${q(as)} ON ${joinCondition}`;
  } else {
    return `\
FROM (
  SELECT *, count(*) OVER () AS ${q('$total')}
  FROM ${table}
  WHERE ${whereCondition}
  ORDER BY ${orderColumnsToString(orderColumns, q)}
  LIMIT ${limit} OFFSET ${offset}
) ${q(as)}`;
  }
}

function orderColumnsToString(orderColumns, q) {
  q = q || doubleQuote;
  const conditions = [];
  for (let column in orderColumns) {
    conditions.push(`${q(column)} ${orderColumns[column]}`);
  }
  return conditions.join(', ');
}

function interpretForOffsetPaging(node, dialect) {
  const { name } = dialect;
  if (node.args && node.args.last) {
    throw new Error('Backward pagination not supported with offsets. Consider using keyset pagination instead');
  }
  const orderColumns = {};
  if (typeof node.orderBy === 'object') {
    for (let column in node.orderBy) {
      let direction = node.orderBy[column].toUpperCase();
      if (direction !== 'ASC' && direction !== 'DESC') {
        throw new Error(direction + ' is not a valid sorting direction');
      }
      orderColumns[column] = direction;
    }
  } else if (typeof node.orderBy === 'string') {
    orderColumns[node.orderBy] = 'ASC';
  } else {
    throw new Error('"orderBy" is required for pagination');
  }
  let limit = ['mariadb', 'mysql'].includes(name) ? '18446744073709551615' : 'ALL';
  let offset = 0;
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      offset = (0, _graphqlRelay.cursorToOffset)(node.args.after) + 1;
    }
  }
  return { limit, offset, orderColumns };
}

function interpretForKeysetPaging(node, dialect) {
  const { name, quote } = dialect;
  const orderColumns = {};
  let descending = node.sortKey.order.toUpperCase() === 'DESC';

  if (node.args && node.args.last) {
    descending = !descending;
  }
  for (let column of (0, _util.wrap)(node.sortKey.key)) {
    orderColumns[column] = descending ? 'DESC' : 'ASC';
  }

  let limit = ['mariadb', 'mysql'].includes(name) ? '18446744073709551615' : 'ALL';
  let whereCondition = '';
  if (node.args && node.args.first) {
    limit = parseInt(node.args.first) + 1;
    if (node.args.after) {
      const cursorObj = (0, _util.cursorToObj)(node.args.after);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending, quote);
    }
    if (node.args.before) {
      throw new Error('Using "before" with "first" is nonsensical.');
    }
  } else if (node.args && node.args.last) {
    limit = parseInt(node.args.last) + 1;
    if (node.args.before) {
      const cursorObj = (0, _util.cursorToObj)(node.args.before);
      validateCursor(cursorObj, (0, _util.wrap)(node.sortKey.key));
      whereCondition = sortKeyToWhereCondition(cursorObj, descending, quote);
    }
    if (node.args.after) {
      throw new Error('Using "after" with "last" is nonsensical.');
    }
  }

  return { limit, orderColumns, whereCondition };
}

function validateCursor(cursorObj, expectedKeys) {
  const actualKeys = Object.keys(cursorObj);
  const expectedKeySet = new Set(expectedKeys);
  const actualKeySet = new Set(actualKeys);
  for (let key of actualKeys) {
    if (!expectedKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the sort key.`);
    }
  }
  for (let key of expectedKeys) {
    if (!actualKeySet.has(key)) {
      throw new Error(`Invalid cursor. The column "${key}" is not in the cursor.`);
    }
  }
}

function sortKeyToWhereCondition(keyObj, descending, q) {
  const sortColumns = [];
  const sortValues = [];
  for (let key in keyObj) {
    sortColumns.push(`${q(key)}`);
    sortValues.push((0, _util.maybeQuote)(keyObj[key]));
  }
  const operator = descending ? '<' : '>';
  return `(${sortColumns.join(', ')}) ${operator} (${sortValues.join(', ')})`;
}